/**
   Wif_IR_Remote_ESP8266
   https://www.forward.com.au/pfod/HomeAutomation/Wifi_IR_Remote/index.html

   (c)2017 Forward Computing and Control Pty. Ltd.
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/

/*
  WiFiConfg is an addon version of An ESP8266 Cheap and Simple Wifi Shield for Arduino and other micros
  This file has been modified to fix the configured baud rate to 115200 to match IR_Remote_TeensyLC.ino

  To Config
  Short the config link and apply power
    and connect to the WifiWebConfig access point
  using the password set in  #define WifiWebConfigPASSWORD  (below)
  Scan this password from the QR code generated with http://www.forward.com.au/pfod/secureChallengeResponse/keyGenerator/index.html

  Once you connect to the Access Point open the webpage
  http://10.1.1.1 to setup your network parameters
  ssid, password and either DHCP or staticIP and portNo.
  Then when you reboot (with the config link removed) you can connect to the ip:portNo and pass data via the UART in both directions.

  The micro board that is connected to this module should have setup() code stating with

  void setup() {
   delay(1000);  // skip skip the ESP8266 debug output
   Serial.begin(19200); // or whatever baud rate you configured via the webpage

    .... rest of setup here
     ..


  see http://www.forward.com.au/pfod/pfodWifiConfig/ESP8266/pfodWifiConfig_ESP8266.html for details
  For an example QR code image look in the directory this file is in.
*/

/**
    Cheap and Simple Wifi Shield for Arduino and other micros
   http://www.forward.com.au/pfod/CheapWifiSheild/index.html

   (c)2015 Forward Computing and Control Pty. Ltd.
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/
#include "WifiWebConfig.h" // include bufferedClient and server and client vars
#include "ESP8266_ledFlasher.h" // include vars for Led flasher

//---------- Arduino 1.6.6+ needs these declarations -------------
void handleNotFound();
void handleRoot();
void handleConfig();
void setupAP(const char* ssid_wifi, const char* password_wifi);


// normally DEBUG is commented out
#define DEBUG

bool FORCE_CONFIG_MODE = false; // set this to true to force webpage config mode ALWAYS, used for testing
// if the next line is uncommented then the buad rate option is removed from the web page and
// the baud rate alwasy set to FORCE_BAUD_RATE
// must be >= 0,  use 0 to disable Serial
#define FORCE_BAUD_RATE 115200

// uncomment this to connect / disconnect messages on Serial out.
//#define CONNECTION_MESSAGES

// the default to show in the config web page
const int DEFAULT_CONNECTION_TIMEOUT_SEC = 10;

// =============== start of WifiWebConfig settings ==============
// update this define with the password from your QR code
//  http://www.forward.com.au/pfod/secureChallengeResponse/keyGenerator/index.html
#define WifiWebConfigPASSWORD ""
//#define WifiWebConfigPASSWORD "b0Ux9akSiwKkwCtcnjTnpWp"
#define WifiWebConfigAP "Wifi_IR_Remote_Config"

// note pfodSecurity uses 19 bytes of eeprom usually starting from 0 so
// start the eeprom address from 20 for configureWifiConfig
const uint8_t webConfigEEPROMStartAddress = 20;

int wifiSetup_pin = 12; // name the input pin for setup mode detection GPIO2 on most ESP8266 boards
int wifiSetup_Gnd_pin = 13; // the pin providing the GND for wifiSetup_pin Set this to <0 if no gnd pin needed
// i.e. int wifiSetup_Gnd_pin = -1;  assumes your switch connects wifiSetup_pin to a Gnd point

// =============== end of WifiWebConfig settings ==============

// On ESP8266-01 and Adafruit HAZZAH ESP8266, connect LED + 270ohm resistor from D2 (GPIO2) to +3V3 to indicate when in config mode

// set the EEPROM structure
struct EEPROM_storage {
  uint32_t timeout;
  uint32_t baudRate;
  uint16_t portNo;
  char ssid[pfodESP8266Utils::MAX_SSID_LEN + 1]; // WIFI ssid + null
  char password[pfodESP8266Utils::MAX_PASSWORD_LEN + 1]; // WiFi password,  if empyt use OPEN, else use AUTO (WEP/WPA/WPA2) + null
  char staticIP[pfodESP8266Utils::MAX_STATICIP_LEN + 1]; // staticIP, if empty use DHCP + null
} storage;
const int EEPROM_storageSize = sizeof(EEPROM_storage);

char ssid_found[pfodESP8266Utils::MAX_SSID_LEN + 1]; // max 32 chars + null
ESP8266WebServer webconfigServer(80);  // this just sets portNo nothing else happens until begin() is called

byte inConfigMode = 0; // false
uint32_t timeout = 0;
const char *aps;

// returns configured connection timeout
uint32_t getConnectionTimeout() {
  return timeout;
}

// returns true if in config mode else false
// if returns true skip reset of setup
bool configSetup ( void ) {
  initLedFlasher();
  WiFi.mode(WIFI_STA);
  inConfigMode = 0; // not in config mode
  EEPROM.begin(512); // must be greater than (wifiConfigStartAddress + EEPROM_storageSize)
  delay(10);
  if (wifiSetup_Gnd_pin >= 0 ) {
    pinMode(wifiSetup_Gnd_pin, OUTPUT);
    digitalWrite(wifiSetup_Gnd_pin, LOW); // make GPIO  low after ESP8266 has initialized
  }

#ifdef DEBUG
  // initial baud rate in DEBUG is 115200
  Serial.begin(115200);
  Serial.println();
  Serial.println(F("Starting Setup"));
  //  bufferedClient.setDebugStream(&Serial);  // add this line if using DEBUG in pfodESP8266BufferedClient library code
#endif

  //============ WifiWebConfig config in Access Point mode ====================
  // see if config button is pressed
  if ((digitalRead(wifiSetup_pin) == LOW) || FORCE_CONFIG_MODE) {
    inConfigMode = 1; // in config mode
    WiFi.mode(WIFI_AP_STA);
#ifdef DEBUG
    Serial.println(F("Setting up Access Point for WifiWebConfig"));
#endif
    // connect to temporary wifi network for setup
    // the features determine the format of the {set...} command
    setupAP(WifiWebConfigAP, WifiWebConfigPASSWORD);
    //   Need to reboot afterwards
    setLedFlashType(ALERT_LED_FLASH); // alert we are in config mode
    return true; // skip rest of setup();
  }
  //============ end WifiWebConfig config ====================

  // else button was not pressed continue to load the stored network settings
  //else use configured setttings from EEPROM
  uint8_t * byteStorageRead = (uint8_t *)&storage;
  for (size_t i = 0; i < EEPROM_storageSize; i++) {
    byteStorageRead[i] = EEPROM.read(webConfigEEPROMStartAddress + i);
  }

  timeout = storage.timeout; // set the timeout

  // Initialise wifi module
#ifdef DEBUG
  Serial.println();
  Serial.println(F("Connecting to AP"));
  Serial.print("ssid '");
  Serial.print(storage.ssid);
  Serial.println("'");
  Serial.print("password '");
  Serial.print(storage.password);
  Serial.println("'");
#endif
  server = WiFiServer(storage.portNo);

  // do this before begin to set static IP and not enable dhcp
  if (*storage.staticIP != '\0') {
    // config static IP
    IPAddress ip(pfodESP8266Utils::ipStrToNum(storage.staticIP));
    IPAddress gateway(ip[0], ip[1], ip[2], 1); // set gatway to ... 1
#ifdef DEBUG
    Serial.print(F("Setting gateway to: "));
    Serial.println(gateway);
#endif
    IPAddress subnet(255, 255, 255, 0);
    WiFi.config(ip, gateway, subnet);
  } // else leave as DHCP

  WiFi.begin(storage.ssid, storage.password);
  int countC = 0;
  while ((WiFi.status() != WL_CONNECTED) && (countC < 12)) { // max 6 sec.
    delay(500);
#ifdef DEBUG
    Serial.print(".");
#endif
    countC++;
    setLedFlashType(FAST_LED_FLASH);
  }
  if (WiFi.status() != WL_CONNECTED) {
#ifdef DEBUG
    Serial.println();   Serial.println(F("Could not connect to Network!"));
#endif
    return false;
  }

#ifdef DEBUG
  Serial.println();  Serial.println(F("Connected to Network!"));
#endif
  networkConnected = true;
  setLedFlashType(NO_LED_FLASH);


  // Start listening for connections
#ifdef DEBUG
  Serial.println(F("Start Server"));
#endif
  server.begin();
  server.setNoDelay(true); // does not do much if anything
#ifdef DEBUG
  Serial.println(F("Server Started"));
  // Print the IP address
  Serial.print(WiFi.localIP());
  Serial.print(':');
  Serial.println(storage.portNo);
  Serial.println(F("Listening for connections..."));
#endif

#ifdef FORCE_BAUD_RATE
  // override storage with the forced baud rage
  storage.baudRate = FORCE_BAUD_RATE;
#endif

#ifdef DEBUG
  Serial.println("End of Setup()"); // end of setup start listening now
  if (storage.baudRate > 0) {
    Serial.print(F("Restarting Serial with configured baud rate:"));
    Serial.println(storage.baudRate);
  } else {
    Serial.print(F("Serial baud rate <=0 Stop Serial and do not restart"));
  }
  delay(100);
#endif
  // restart Serial with configured baudrate
  Serial.end();
  delay(10);
  if (storage.baudRate > 0) {
    // start it again else do not start
    Serial.begin(storage.baudRate);
    delay(10);
  }

  // initialize client
  client = server.available(); // evaluates to false if no connection

  return false;  // not in config mode continue with rest of main setup
}


// call this from main loop() if returns true then just return from main loop
bool configLoop() {
  if (inConfigMode) {
    webconfigServer.handleClient();
    delay(0);
    return true;  // return from main loop() now
  }
  // else
  return false; // continue with main loop()
}



void setupAP(const char* ssid_wifi, const char* password_wifi) {
  aps = pfodESP8266Utils::scanForStrongestAP((char*)&ssid_found, pfodESP8266Utils::MAX_SSID_LEN + 1);
  delay(0);
  IPAddress local_ip = IPAddress(10, 1, 1, 1);
  IPAddress gateway_ip = IPAddress(10, 1, 1, 1);
  IPAddress subnet_ip = IPAddress(255, 255, 255, 0);
#ifdef DEBUG
  Serial.println(F("configure WifiWebConfig"));
#endif
  WiFi.softAP(ssid_wifi, password_wifi);

#ifdef DEBUG
  Serial.println(F("Access Point setup"));
#endif
  WiFi.softAPConfig(local_ip, gateway_ip, subnet_ip);

#ifdef DEBUG
  Serial.println("done");
  IPAddress myIP = WiFi.softAPIP();
  Serial.print(F("AP IP address: "));
  Serial.println(myIP);
#endif
  delay(10);
  webconfigServer.on ( "/", handleRoot );
  webconfigServer.on ( "/config", handleConfig );
  webconfigServer.onNotFound ( handleNotFound );
  webconfigServer.begin();
#ifdef DEBUG
  Serial.println ( "HTTP webconfigServer started" );
#endif
}

void handleConfig() {
  // set defaults
  uint16_t portNo = 80;
  uint32_t timeout = DEFAULT_CONNECTION_TIMEOUT_SEC * 1000; // time out in mS
  uint32_t baudRate = 19200;
#ifdef FORCE_BAUD_RATE
  baudRate = FORCE_BAUD_RATE;
#endif
  storage.baudRate = (uint32_t)(baudRate); // set a default

  char tempStr[20]; // for rate,port,timeout

  if (webconfigServer.args() > 0) {
#ifdef DEBUG
    String message = "Config results\n\n";
    message += "URI: ";
    message += webconfigServer.uri();
    message += "\nMethod: ";
    message += ( webconfigServer.method() == HTTP_GET ) ? "GET" : "POST";
    message += "\nArguments: ";
    message += webconfigServer.args();
    message += "\n";
    for ( uint8_t i = 0; i < webconfigServer.args(); i++ ) {
      message += " " + webconfigServer.argName ( i ) + ": " + webconfigServer.arg ( i ) + "\n";
    }
    Serial.println(message);
    Serial.println();
#endif

    uint8_t numOfArgs = webconfigServer.args();
    const char *strPtr;
    uint8_t i = 0;
    for (; (i < numOfArgs); i++ ) {
      // check field numbers
      if (webconfigServer.argName(i)[0] == '1') {
        pfodESP8266Utils::strncpy_safe(storage.ssid, (webconfigServer.arg(i)).c_str(), pfodESP8266Utils::MAX_SSID_LEN);
        pfodESP8266Utils::urldecode2(storage.ssid, storage.ssid); // result is always <= source so just copy over

      } else if (webconfigServer.argName(i)[0] == '2') {
        pfodESP8266Utils::strncpy_safe(storage.password, (webconfigServer.arg(i)).c_str(), pfodESP8266Utils::MAX_PASSWORD_LEN);
        pfodESP8266Utils::urldecode2(storage.password, storage.password); // result is always <= source so just copy over
        // if password all blanks make it empty
        if (pfodESP8266Utils::isEmpty(storage.password)) {
          storage.password[0] = '\0';
        }

      } else if (webconfigServer.argName(i)[0] == '3') {
        pfodESP8266Utils::strncpy_safe(storage.staticIP, (webconfigServer.arg(i)).c_str(), pfodESP8266Utils::MAX_STATICIP_LEN);
        pfodESP8266Utils::urldecode2(storage.staticIP, storage.staticIP); // result is always <= source so just copy over
        if (pfodESP8266Utils::isEmpty(storage.staticIP)) {
          // use dhcp
          storage.staticIP[0] = '\0';
        }

      } else if (webconfigServer.argName(i)[0] == '4') {
        // convert portNo to uint16_6
        pfodESP8266Utils::strncpy_safe(tempStr, (webconfigServer.arg(i)).c_str(), sizeof(tempStr));
#ifdef DEBUG
        Serial.print(F("portNoStr:"));
        Serial.println(tempStr);
#endif
        long longPort = 0;
        pfodESP8266Utils::parseLong((byte*)tempStr, &longPort);
        storage.portNo = (uint16_t)longPort;

      } else if (webconfigServer.argName(i)[0] == '5') {
        // convert baud rate to int32_t
        pfodESP8266Utils::strncpy_safe(tempStr, (webconfigServer.arg(i)).c_str(), sizeof(tempStr));
#ifdef DEBUG
        Serial.print(F("baudStr:"));
        Serial.println(tempStr);
#endif
        long baud = 0;
        pfodESP8266Utils::parseLong((byte*)tempStr, &baud);
        storage.baudRate = (uint32_t)(baud);

      } else if (webconfigServer.argName(i)[0] == '6') {
        // convert timeout to int32_t
        // then *1000 to make mS and store as uint32_t
        pfodESP8266Utils::strncpy_safe(tempStr, (webconfigServer.arg(i)).c_str(), sizeof(tempStr));
#ifdef DEBUG
        Serial.print(F("timeoutStr:"));
        Serial.println(tempStr);
#endif

        long timeoutSec = 0;
        pfodESP8266Utils::parseLong((byte*)tempStr, &timeoutSec);
        if (timeoutSec > 4294967) {
          timeoutSec = 4294967;
        }
        if (timeoutSec < 0) {
          timeoutSec = 0;
        }
        storage.timeout = (uint32_t)(timeoutSec * 1000);
      }
    }
#ifdef DEBUG
    Serial.println();
    printStorage(&storage);
#endif
    uint8_t * byteStorage = (uint8_t *)&storage;
    for (size_t i = 0; i < EEPROM_storageSize; i++) {
      EEPROM.write(webConfigEEPROMStartAddress + i, byteStorage[i]);
    }
    delay(0);
    bool committed =  EEPROM.commit();
#ifdef DEBUG
    Serial.print(F("EEPROM.commit() = "));
    Serial.println(committed ? "true" : "false");
#endif

  } // else if no args just return current settings

  delay(0);
  struct EEPROM_storage storageRead;
  uint8_t * byteStorageRead = (uint8_t *)&storageRead;
  for (size_t i = 0; i < EEPROM_storageSize; i++) {
    byteStorageRead[i] = EEPROM.read(webConfigEEPROMStartAddress + i);
  }
#ifdef DEBUG
  Serial.println();
  printStorage(&storageRead);
#endif
  String rtnMsg = "<html>"
                  "<head>"
                  "<title>WifiWebConfig Server Setup</title>"
                  "<meta charset=\"utf-8\" />"
                  "<meta name=viewport content=\"width=device-width, initial-scale=1\">"
                  "</head>"
                  "<body>"
                  "<h2>WifiWebConfig Server Settings saved.</h2><br>Power cycle to connect to ";
  if (storageRead.password[0] == '\0') {
    rtnMsg += "the open network ";
  }
  rtnMsg += "<b>";
  rtnMsg += storageRead.ssid;
  rtnMsg += "</b>";

  if (storageRead.staticIP[0] == '\0') {
    rtnMsg += "<br> using DCHP to get its IP address";
  } else { // staticIP
    rtnMsg += "<br> using IP addess ";
    rtnMsg += "<b>";
    rtnMsg += storageRead.staticIP;
    rtnMsg += "</b>";
  }
  rtnMsg += "<br><br>Will start a server listening on port ";
  rtnMsg += storageRead.portNo;
  rtnMsg += ".<br> with connection timeout of ";
  rtnMsg += storageRead.timeout / 1000;
  rtnMsg += " seconds.";
  rtnMsg += "<br><br>Serial baud rate set to ";
  rtnMsg += storageRead.baudRate;
  "</body>"
  "</html>";

  webconfigServer.send ( 200, "text/html", rtnMsg );
}


void handleRoot() {
  String msg;
  msg = "<html>"
        "<head>"
        "<title>WifiWebConfig Server Setup</title>"
        "<meta charset=\"utf-8\" />"
        "<meta name=viewport content=\"width=device-width, initial-scale=1\">"
        "</head>"
        "<body>"
        "<h2>WifiWebConfig Server Setup</h2>"
        "<p>Use this form to configure your device to connect to your Wifi network and start as a Server listening on the specified port.</p>"
        "<form class=\"form\" method=\"post\" action=\"/config\" >"
        "<p class=\"name\">"
        "<label for=\"name\">Network SSID</label><br>"
        "<input type=\"text\" name=\"1\" id=\"ssid\" placeholder=\"wifi network name\"  required "; // field 1

  if (*aps != '\0') {
    msg += " value=\"";
    msg += aps;
    msg += "\" ";
  }
  msg += " />"
         "<p class=\"password\">"
         "<label for=\"password\">Password for WEP/WPA/WPA2 (enter a space if there is no password, i.e. OPEN)</label><br>"
         "<input type=\"text\" name=\"2\" id=\"password\" placeholder=\"wifi network password\" autocomplete=\"off\" required " // field 2
         "</p>"
         "<p class=\"static_ip\">"
         "<label for=\"static_ip\">Set the Static IP for this device</label><br>"
         "(If this field is empty, DHCP will be used to get an IP address)<br>"
         "<input type=\"text\" name=\"3\" id=\"static_ip\" placeholder=\"192.168.4.99\" "  // field 3
         " pattern=\"\\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\b\"/>"
         "</p>"
         "<p class=\"portNo\">"
         "<label for=\"portNo\">Set the port number that the Server will listen on for connections.</label><br>"
         "<input type=\"text\" name=\"4\" id=\"portNo\" placeholder=\"80\" required"  // field 4
         " pattern=\"\\b([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])\\b\" />"
         "</p>";
#ifndef FORCE_BAUD_RATE
  // if FORCE_BAUD_RATE not defined prompt to set it
  msg +=         "<p class=\"baud\">"
                 "<label for=\"baud\">Serial Baud Rate (limit to 19200 for Uno and Mega)</label><br>"
                 "<select name=\"5\" id=\"baud\" required>" // field 5
                 "<option value=\"9600\">9600</option>"
                 "<option value=\"14400\">14400</option>"
                 "<option selected value=\"19200\">19200</option>"
                 "<option value=\"28800\">28800</option>"
                 "<option value=\"38400\">38400</option>"
                 "<option value=\"57600\">57600</option>"
                 "<option value=\"76800\">76800</option>"
                 "<option value=\"115200\">115200</option>"
                 "</select>"
                 "</p>";
#endif // FORCE_BAUD_RATE
  msg +=          "<p class=\"timeout\">"
                  "<label for=\"timeout\">Set the server connection timeout in seconds, 0 for never timeout (not recommended).</label><br>"
                  "<input type=\"text\" name=\"6\" id=\"timeout\" required"  // field 6
                  " value=\"";
  msg +=   DEFAULT_CONNECTION_TIMEOUT_SEC;
  msg +=  "\""
          " pattern=\"\\b([0-9]{1,7})\\b\" />"
          "</p>"
          "<p class=\"submit\">"
          "<input type=\"submit\" value=\"Configure\"  />"
          "</p>"
          "</form>"
          "</body>"
          "</html>";

  webconfigServer.send ( 200, "text/html", msg );
}


void handleNotFound() {
  handleRoot();
}

void printStorage(struct EEPROM_storage* storagePtr) {
#ifdef DEBUG
  Serial.print("timeout:");
  Serial.println(storagePtr->timeout);
  Serial.print("baudRate:");
  Serial.println(storagePtr->baudRate);
  Serial.print("portNo:");
  Serial.println(storagePtr->portNo);
  Serial.print("ssid:");
  Serial.println(storagePtr->ssid);
  Serial.print("password:");
  Serial.println(storagePtr->password);
  Serial.print("staticIP:");
  Serial.println(storagePtr->staticIP);
#endif
}


